parse(Ws,Tree) :-
    parse(Ws,[],[Tree]),
    Tree = [S|_], start_symbol(S).

parse([W|Ws],Stack,Reduced) :-
    ( reduce(Stack,RedStack),
        parse([W|Ws],RedStack,Reduced)
    ; shift(W,Stack,WStack),
        parse(Ws,WStack,Reduced) ).
parse([],Stack,Reduced) :-
    ( reduce(Stack,RedStack),
        parse([],RedStack,Reduced)
    ; Reduced = Stack ).

shift(W,Stack,[Tree|Stack]) :-
    ( word(W,A), Tree = [A,W]
    ; A = [W], Tree = [A] ).
reduce(Stack,Reduced) :-
    (B ---> Beta),
    roots(Beta,Roots),
    split(Stack,Roots,BetaForest,AlphaForest),
    reverse(BetaForest,BetaTrees),
    Tree = [B|BetaTrees],
    Reduced = [Tree|AlphaForest].
roots((A,B),Cats) :-
    !, roots(A,CatsA), roots(B,CatsB),
    append(CatsA,CatsB,Cats).
roots((A;B),Cats) :-
    !, (roots(A,Cats) ; roots(B,Cats)).
roots([W|Ws],Cats) :-
    !, roots(Ws,CatsWs), Cats = [[W]|CatsWs].
roots([],Cats) :-
    !, Cats = [].
roots({Code},Cats) :-
    !, call(Code), Cats = [].
roots(Cat,[Cat]).
split(Forest,Beta,BetaForest,AlphaForest) :-    
    reverse(Beta,BetaRev),
    newTrees(BetaRev,BetaForest),
    append(BetaForest,AlphaForest,Forest).

newTrees([A|As],[[A|_]|Ts]) :-
    newTrees(As,Ts).
newTrees([],[]).
