:- dynamic changed/1, goto/3, done/1, action/3.

compute_lr0_automaton :-
    remove_old_automata,
    compute_initial_state,    % endl.Automat  A_G
    compute_successor_states, % endl.Automat  A_G
    compute_actions.          % Kellerautomat M_G

remove_old_automata :-        % aufraeumen
    retractall(action(_,_,_)),
    retractall(maxstate(_)),
    retractall(state(_)),
    retractall(state(_,_)),
    retractall(closure(_,_)),
    retractall(goto(_,_,_)),
    retractall(changed(_)),
    retractall(done(_)).

compute_initial_state :-
    assert(state(0)),
    assert(maxstate(0)),
    ( start_symbol(S),
        add(state(0,(start --> [],[S,end]))),
        fail
    ; true ),
    close_state(0).
close_state(P) :-
    ( state(P,Rule),
        add(closure(P,Rule)),
        add(changed(P)),
        fail
    ; (changed(P) -> close_state2(P) ; true)
    ).
close_state2(P) :-
    retract(changed(P)),
    ( closure(P,(A --> Alpha,[C|Beta])),
        (C --> GammaTuple),
        tupleToList(GammaTuple,Gamma),
        \+ known_instance(closure(P,(C --> [],Gamma))),
        add(closure(P,(C --> [],Gamma))),
        add(changed(P)),
        fail
    ; (changed(P) -> close_state2(P) ; true) ).
% ---- Hilfsdefinitionen ----
add(A) :-
    known_instance(A), !.
add(A) :-
    assertz(A).
known_instance(A) :- % ist A[Xi/consti] beweisbar?
        \+ \+ (numbervars(A, const, 0, B), A).

tupleToList((A, B), [A|C]) :- 
    !, tupleToList(B, C).
tupleToList(A, [A]).
% ----------------------------
compute_successor_states :-
    state(P), \+ done(P),
    !, compute_successors(P), add(done(P)),
    compute_successor_states.
compute_successor_states.
    
compute_successors(P) :-
    ( setof(B,A^Alpha^Beta^closure(P,
             (A --> Alpha,[B|Beta])),Bs)
    -> compute_successors(P,Bs)
    ; true ).

compute_successors(P,[B|Bs]) :-
    mk_succ(P,B),
    compute_successors(P,Bs).
compute_successors(P,[]).

mk_succ(P,end) :- !.
mk_succ(P,B) :-
    ( closure(P,(A --> Alpha,[B|Beta])),
        append(Alpha,[B],AlphaB),
        add(state(newstate,(A --> AlphaB,Beta))), 
        fail              % omit: state(newstate)!
    ; close_state(newstate) ),
    number(newstate,Q),
    ( retract(state(newstate,Rule)),
        add(state(Q,Rule)),
        fail
    ; retract(closure(newstate,Rule)),
        add(closure(Q,Rule)),
        fail
    ; true ),
    add(goto(P,B,Q)). 

number(newstate,N) :-
    setof(Q,state(Q),Qs),
    number(newstate,Qs,N).

number(newstate,[Q|Qs],N) :-
    ( same_states(newstate,Q)
    -> N=Q ; number(newstate,Qs,N) ).
number(newstate,[],N) :-
    maxstate(M),
    retract(maxstate(M)),
    N is M+1,
    assert(maxstate(N)),
    add(state(N)).

same_states(Q,P) :-
    setof(Rule,closure(Q,Rule),ClosureQ),
    setof(Rule,closure(P,Rule),ClosureP),
    ClosureQ = ClosureP.  

compute_actions :-
    state(Q),
    terminal(B),
    ( closure(Q,(A --> Alpha,[])),
        add(action(Q,B,reduce(A --> Alpha)))
    ; closure(Q,(_A --> _Alpha,[B|_])),
        goto(Q,B,P),
        add(action(Q,B,shift(P)))
    ; closure(Q,(A --> Alpha,[end])),
        add(action(Q,end,accept))
    ),
    fail.
compute_actions :-
    state(Q),
    terminal(B), 
    (action(Q,B,_) -> true ; add(action(Q,B,error))),
    fail.
compute_actions.

terminal(B) :-
    (_ --> GammaTuple),
    tupleToList(GammaTuple,Gamma),    
    member(B,Gamma),
    \+ (B --> _).

